import os

from data_tools.query.flux import FluxQuery
from data_tools.collections.time_series import TimeSeries
from influxdb_client import InfluxDBClient
from dotenv import load_dotenv
import pandas as pd


INFLUX_URL = "http://influxdb.telemetry.ubcsolar.com"


class DBClient:
    """
    This class encapsulates a connection to an InfluxDB database.
    """
    def __init__(self, influxdb_org=None, influxdb_token=None):
        if influxdb_token is None or influxdb_org is None:
            load_dotenv()

            influxdb_token = os.getenv("INFLUX_TOKEN")
            influxdb_org = os.getenv("INFLUX_ORG")

        print(f"Creating client with API Token: {influxdb_token}")
        print(f"Creating client with Org: {influxdb_org}")

        self._influx_org = influxdb_org

        self._client = InfluxDBClient(url=INFLUX_URL, org=influxdb_org, token=influxdb_token)

    def get_buckets(self) -> list[str]:
        """
        Obtain the available buckets.
        :return: list of available buckets
        """
        result = self._client.buckets_api().find_buckets()
        buckets = [bucket.name for bucket in result.buckets]
        return list(buckets)

    def get_measurements_in_bucket(self, bucket: str) -> list[str]:
        """
        Obtain a list of the possible measurements within a bucket.

        :param bucket: the bucket that will be queried. Must exist.
        :return: a list of available measurements in `bucket`.
        """
        query = f'''
                import "influxdata/influxdb/schema"
                schema.measurements(bucket: "{bucket}")
                '''

        result = self._client.query_api().query(org=self._influx_org, query=query)
        measurements = {record.get_value() for table in result for record in table.records}

        return list(measurements)

    def get_fields_in_measurement(self, bucket: str, measurement: str) -> list[str]:
        """
        Obtain the possible fields within a measurement and bucket.
        :param bucket: the bucket that will be queried. Must exist.
        :param measurement: the measurement that will be queried. Must exist.
        :return:
        """
        field_query = f'''from(bucket: "{bucket}")
                      |> range(start: -30d)
                      |> filter(fn: (r) => r["_measurement"] == "{measurement}")
                      |> group(columns: ["_field"])
                      |> distinct(column: "_field")'''

        field_result = self._client.query_api().query(org=self._influx_org, query=field_query)
        fields = {record.get_value() for table in field_result for record in table.records}

        return list(fields)

    def get_cars_in_measurement(self, bucket: str, measurement: str) -> list[str]:
        """
        Obtain the different cars within a measurement and bucket.
        :param bucket: the bucket that will be queried. Must exist.
        :param measurement: the measurement that will be queried. Must exist.
        :return:
        """
        field_query = f'''from(bucket: "{bucket}")
                      |> range(start: -30d)
                      |> distinct(column: "car")'''

        cars_result = self._client.query_api().query(org=self._influx_org, query=field_query)
        cars = {record.get_value() for table in cars_result for record in table.records}

        return list(cars)

    def query_dataframe(self, query: FluxQuery) -> pd.DataFrame:
        """
        Submit a Flux query, and return the result as a DataFrame.

        :param FluxQuery query: the query which will be submitted
        :return: the resulting data as a DataFrame
        """
        compiled_query = query.compile_query()
        compiled_query += ' |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value") '

        return self._client.query_api().query_data_frame(compiled_query)

    def query_time_series(self, start: str, stop: str, field: str, bucket: str = "CAN_log", car: str = "Brightside", granularity: float = 0.1, units: str = "", measurement: str = None) -> TimeSeries:
        """
        Query the database for a specific field, over a certain time range.
        The data will be processed into a TimeSeries, which has homogenous and evenly-spaced (temporally) elements.
        Data will be re-interpolated to have temporal granularity of ``granularity``.

        :param start: the start time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param stop: the end time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param field: the field which is to be queried.
        :param str bucket: the bucket which will be queried
        :param car: the car which data is being queried for, default is "Brightside".
        :param granularity: the temporal granularity of the resulting TimeSeries in seconds, default is 0.1s.
        :param units: the units of the returned data, optional.
        :return: a TimeSeries of the resulting time-series data
        """
        # Make the query
        query = FluxQuery().from_bucket("CAN_log").range(start=start, stop=stop).filter(field=field).car(car)
        if measurement:
            query = query.filter(measurement=measurement)
        query_df = self.query_dataframe(query)

        if isinstance(query_df, list):
            raise ValueError("Query returned multiple fields! Please refine your query.")

        if len(query_df) == 0:
            raise ValueError("Query is empty! Verify that the data is visible on InfluxDB for the queried bucket.")

        return TimeSeries.from_query_dataframe(query_df, granularity, field, units)


if __name__ == "__main__":
    start = "2024-07-07T02:23:57Z"
    stop = "2024-07-07T02:34:15Z"
    client = DBClient()

    pack_voltage: TimeSeries = client.query_time_series(start, stop, "MotorCurrent", units="V", measurement="MCB")
    pack_current: TimeSeries = client.query_time_series(start, stop, "PackCurrent", units="A")
    vehicle_velocity: TimeSeries = client.query_time_series(start, stop, "VehicleVelocity", units="m/s")
    pack_current, pack_voltage = TimeSeries.align(pack_current, pack_voltage)





"""
===========================================
Querying Tools (:mod:`data_tools.query`)
===========================================

Flux Tools
==========

.. autosummary::
   :toctree: generated/

   FluxStatement      -- Atomic component of FluxQuery
   FluxQuery          -- Query composed of FluxStatements

InfluxDB Tools
==============

.. autosummary::
   :toctree: generated/

   DBClient           -- Powerful and simple InfluxDB client

PostgreSQL Tools
================

.. autosummary::
   :toctree: generated/

   PostgresClient    -- Powerful and simple InfluxDB client

"""


from .flux import FluxQuery, FluxStatement
from .influxdb_query import DBClient
from .postgresql_query import PostgresClient
from .data_schema import get_sensor_id, get_data_units, CANLog, init_schema


__all__ = [
    "FluxQuery",
    "FluxStatement",
    "DBClient",
    "PostgresClient"
]



from datetime import datetime, timezone
from sqlalchemy import create_engine, Engine
from sqlalchemy.orm import sessionmaker, Session
from data_tools.query.data_schema import init_schema, CANLog, get_sensor_id, get_data_units
from data_tools.collections.time_series import TimeSeries
import matplotlib.pyplot as plt
import numpy as np
from typing import List, Type, Union
import os
from dotenv import load_dotenv

load_dotenv()

POSTGRESQL_USERNAME = os.getenv("POSTGRESQL_USERNAME")
POSTGRESQL_PASSWORD = os.getenv("POSTGRESQL_PASSWORD")
POSTGRESQL_DATABASE = os.getenv("POSTGRESQL_DATABASE")
POSTGRESQL_ADDRESS  = os.getenv("POSTGRESQL_ADDRESS")


def _get_db_url(db_name: str, ip_address: str, username: str, password: str) -> str:
    """
    Get the URL to a Postgres database.

    :param str db_name: the name of the database being connected to
    :param str ip_address: the IP Address of the machine running the PostgreSQL instance being connected to
    :param str username: the username of the user that is connecting
    :param str password: the password of the user that is connecting
    :return: the URL formatted as a string
    """
    assert isinstance(db_name, Union[str, None]), "db_name must be a string!"
    assert isinstance(username, Union[str, None]), "username must be a string!"
    assert isinstance(ip_address, Union[str, None]), "ip_address must be a string!"
    assert isinstance(password, Union[str, None]), "password must be a string!"

    return f"postgresql://{username}:{password}@{ip_address}:5432/{db_name}"


class PostgresClient:
    """

    Connect to a PostgresSQL database and concisely make queries for time-series data.

    """
    def __init__(self, db_name: str, ip_address: str = None, username: str = None, password: str = None):
        assert isinstance(db_name, Union[str, None]), "db_name must be a string!"
        assert isinstance(username, Union[str, None]), "username must be a string!"
        assert isinstance(ip_address, Union[str, None]), "ip_address must be a string!"
        assert isinstance(password, Union[str, None]), "password must be a string!"

        if username is None:
            username = POSTGRESQL_USERNAME
        if ip_address is None:
            ip_address = POSTGRESQL_ADDRESS
        if password is None:
            password = POSTGRESQL_PASSWORD

        url = _get_db_url(db_name, ip_address, username, password)
        self._engine: Engine = create_engine(url)

        self._session_builder = sessionmaker(bind=self._engine)
        self._session: Session = self._session_builder()

    def query(self, field: str, start_time: datetime, end_time: datetime, granularity: float = 1.0,) -> TimeSeries:
        """
        Query the database for time-series data matching ``field``, between ``start_time`` and ``end_time``.

        Undefined behaviour when timestamps provided are not in UTC!

        :param str field: name of the field that will be queried
        :param datetime.datetime start_time: the UTC datetime of the beginning of the data to be queried
        :param datetime.datetime end_time: the UTC datetime of the end of the data to be queried
        :param float granularity: the desired temporal granularity of the returned data
        :return: successfully queried data formatted as a TimeSeries
        :raises ValueError: if no data could be queried
        """
        unix_start_time = start_time.timestamp()
        unix_end_time = end_time.timestamp()

        queried_data: List[Type[CANLog]] = self._session.query(CANLog)\
            .filter(CANLog.sensor_type == get_sensor_id(field),
                    CANLog.timestamp >= unix_start_time,
                    CANLog.timestamp <= unix_end_time)\
            .order_by(CANLog.timestamp).all()

        if len(queried_data) == 0:
            raise ValueError("Query returned no data!")

        timestamps = np.fromiter([datum.timestamp for datum in queried_data], dtype=float)
        values = np.fromiter([datum.value for datum in queried_data], dtype=float)

        # The start and end times of the returned data may not be the same as what was requested,
        # and we need the actual values to store the metadata properly
        actual_start_time = datetime.fromtimestamp(timestamps[0], tz=timezone.utc)
        actual_end_time = datetime.fromtimestamp(timestamps[-1], tz=timezone.utc)

        # Subtract off the first timestamp so every timestamp is delta-time from the first.
        timestamps -= timestamps[0]

        plt.plot(timestamps, values)
        plt.show()

        # Reform the x-axis as a homogenous array
        x_axis = np.arange(timestamps[0], timestamps[-1], granularity)

        # Interpolate the data array onto this new x-axis
        wave = np.interp(x_axis, timestamps, values)

        time_series_dict = {
            "start": actual_start_time,
            "stop": actual_end_time,
            "car": "N/A",
            "measurement": "N/A",
            "field": field,
            "granularity": granularity,
            "length": len(x_axis),
            "units": get_data_units(get_sensor_id(field)),
        }

        return TimeSeries(wave, time_series_dict)

    def init_schema(self):
        """
        Initialize the Postgres database schema. This only needs to be called ONCE PER DATABASE.

        Has no effect if called on an already initialized database.
        """
        init_schema(self._engine)

    def get_session(self) -> Session:
        """
        Obtain a new Session to make queries to this database.
        :return: a new Session bound to this connection
        """
        return self._session_builder()

    def write(self, session: Session, instance: Union[object, List[object]]) -> Union[None, Exception]:
        """
        Write some data ``instance`` to this database using ``session``.

        :param session: the Session that will be used to write to the database
        :param instance: the data that will be uploaded, can be a single object or an iterable of objects.
        :return: None if the data was uploaded successfully, or the Exception that was raised otherwise
        """
        try:
            # Upload data as a batch if it's a list
            if isinstance(data, List):
                session.add_all(instance)
            else:
                session.add(instance)

            session.commit()

        except Exception as e:
            session.rollback()
            return e

        
if __name__ == "__main__":
    field = "VehicleVelocity"
    client = PostgresClient(POSTGRESQL_DATABASE, POSTGRESQL_ADDRESS, POSTGRESQL_USERNAME, POSTGRESQL_PASSWORD)

    start_time = datetime(2024, 7, 16, 15, 0, 0, tzinfo=timezone.utc)
    end_time = datetime(2024, 7, 16, 22, 0, 0, tzinfo=timezone.utc)

    data: TimeSeries = client.query(field, start_time, end_time)

    data.plot()



from .query import (
    FluxQuery,
    FluxStatement,
    DBClient,
    PostgresClient
)

from .collections import (
    TimeSeries,
    FSGPDayLaps
)


__all__ = [
    "FluxQuery",
    "FluxStatement",
    "TimeSeries",
    "DBClient",
    "FSGPDayLaps",
    "PostgresClient"
]
